----------------------------------------------------------------
-- GF 2**m package body (GF2m.adb)
-- 
-- Defines several funtions for finite fields operations
-- of Chapter 7 examples
----------------------------------------------------------------
package body GF2m is   

   function Product_alpha_A(A, F: poly_vector) return poly_vector is
      C: poly_vector;
   begin
      C(0) := m2and(A(m-1),F(0));
      for i in 1 .. m-1 loop C(i) := m2xor(A(i-1),m2and(A(m-1),F(i))); end loop;
      return C;
   end Product_alpha_A; 

---------------------------------------------------------------------------------

function Product_alpha_Am(A, F: poly_vectorm) return poly_vectorm is
      C: poly_vectorm;
   begin
      C(0) := m2and(A(m),F(0));
      for i in 1 .. m loop C(i) := m2xor(A(i-1),m2and(A(m),F(i))); end loop;
      return C;
   end Product_alpha_Am; 

---------------------------------------------------------------------------------

   function Product_alpha(F: poly_vector) return poly_vector is
      C: poly_vector;
   begin
      C(0) := m2and(F(m-1),F(0));
      for i in 1 .. m-1 loop C(i) := m2xor(F(i-1),m2and(F(m-1),F(i))); end loop;
      return C;
   end Product_alpha; 

---------------------------------------------------------------------------------
   
   function m2abv(x: bit; y: poly_vector) return poly_vector is
      z : poly_vector;
   begin
      for i in 0 .. m-1 loop
         z(i) := m2and(x,y(i));
      end loop;
      return z;
   end m2abv;
   
---------------------------------------------------------------------------------
   
   function m2xvv(x: poly_vector; y: poly_vector) return poly_vector is
      z : poly_vector;
   begin
      for i in 0 .. m-1 loop
         z(i) := m2xor(x(i),y(i));
      end loop;
      return z;
   end m2xvv;

---------------------------------------------------------------------------------
   
   function m2xvvm(x: poly_vectorm; y: poly_vectorm) return poly_vectorm is
      z : poly_vectorm;
   begin
      for i in 0 .. m loop
         z(i) := m2xor(x(i),y(i));
      end loop;
      return z;
   end m2xvvm;

---------------------------------------------------------------------------------

   function m2xvv2(x: poly2_vector; y: poly_vector) return poly2_vector is
      z : poly2_vector;
   begin
      for i in 0 .. m-1 loop
         z(i) := m2xor(x(i),y(i));
      end loop;
      for i in m .. 2*m-2 loop
         z(i) := x(i);
      end loop;
      return z;
   end m2xvv2;
   
---------------------------------------------------------------------------------

   function lshift(x: poly_vector) return poly_vector is
      z: poly_vector;
   begin
      for i in 0 .. m-2 loop
         z(i) := x(i+1);
      end loop;
      z(m-1) := 0;
      return z;
   end lshift;

--------------------------------------------------------------------------------- 

   function rshift(x: poly_vector) return poly_vector is
      z: poly_vector;
   begin
      for i in 1 .. m-1 loop
         z(i) := x(i-1);
      end loop;
      z(0) := 0;
      return z;
   end rshift;

--------------------------------------------------------------------------------- 

   function lshiftm(x: poly_vectorm) return poly_vectorm is
      z: poly_vectorm;
   begin
      for i in 0 .. m-1 loop
         z(i) := x(i+1);
      end loop;
      z(m) := 0;
      return z;
   end lshiftm;

--------------------------------------------------------------------------------- 

   function rshiftm(x: poly_vectorm) return poly_vectorm is
      z: poly_vectorm;
   begin
      for i in 1 .. m loop
         z(i) := x(i-1);
      end loop;
      z(0) := 0;
      return z;
   end rshiftm;

--------------------------------------------------------------------------------- 

   function lshift2(x: poly2_vector) return poly2_vector is
      z: poly2_vector;
   begin
      for i in 0 .. 2*m-3 loop
         z(i) := x(i+1);
      end loop;
      z(2*m-2) := 0;
      return z;
   end lshift2;

--------------------------------------------------------------------------------- 

   function poly_multiplication(A,B: poly_vector) return poly2_vector is
      C: poly2_vector;
   begin
      for i in 0 .. 2*m-2 loop C(i) := 0; end loop;

      for k in 0 .. m-1 loop
         for i in 0 .. k loop
            C(k) := m2xor(C(k),m2and(A(i),B(k-i)));
         end loop;
      end loop;
   
      for k in m .. 2*m-2 loop
         for i in k .. 2*m-2 loop
            C(k) := m2xor(C(k),m2and(A(k-i+(m-1)),B(i-(m-1))));
         end loop;
      end loop;
   
      return C;
      
   end poly_multiplication;

---------------------------------------------------------------------------------

   function reduction_matrix_R (f: poly_vector) return poly_matrix_m1m2 is
      R: poly_matrix_m1m2;
   begin
      for j in 0 .. m-1 loop 
         for i in 0 .. m-2 loop
            R(j,i) := 0; 
         end loop;
      end loop;
      
      for j in 0 .. m-1 loop
         R(j,0) := f(j);
      end loop;
 
      for i in 1 .. m-2 loop
         for j in 0 .. m-1 loop
            if j = 0 then 
               R(j,i) := m2and(R(m-1,i-1),R(j,0));
            else
               R(j,i) := m2xor(R(j-1,i-1),m2and(R(m-1,i-1),R(j,0))); 
            end if;
         end loop;
      end loop;
      
      return R;
      
   end reduction_matrix_R;

---------------------------------------------------------------------------------

function matrix_P (f: poly_vector) return poly_matrix_m2m1 is
      P: poly_matrix_m2m1;
   begin
      for j in 0 .. m-1 loop
         P(0,j) := f(j);
      end loop;
   
      for i in 1 .. m-2 loop 
         for j in 0 .. m-1 loop
            P(i,j) := 0; 
         end loop;
      end loop;
    
      for i in 1 .. m-2 loop
         for j in 0 .. m-1 loop
            if j = 0 then 
               P(i,j) := P(i-1,m-1);
            else
               P(i,j) := m2xor(P(i-1,j-1),m2and(P(i-1,m-1),P(0,j))); 
            end if;
         end loop;
      end loop;
      
      return P;
      
   end matrix_P;
   
---------------------------------------------------------------------------------

function mastrovito_matrix (a: poly_vector; P: poly_matrix_m2m1) return poly_matrix is
      Z: poly_matrix;
   begin
      for i in 0 .. m-1 loop
         Z(i,0) := a(i);
      end loop;
   
      for i in 0 .. m-1 loop 
         for j in 1 .. m-1 loop
            Z(i,j) := 0; 
         end loop;
      end loop;
    
      for i in 0 .. m-1 loop
         for j in 1 .. m-1 loop
            for t in 0 .. j-1 loop
               Z(i,j) := m2xor(Z(i,j),m2and(P(j-1-t,i),a(m-1-t)));
            end loop;
            if i >= j then   
               Z(i,j) := m2xor(a(i-j),Z(i,j));        
            end if;
         end loop;
      end loop;
      
      return Z;
      
   end mastrovito_matrix;
   
--------------------------------------------------------------------------------- 

   function vector_D(a,b: poly_vector) return poly_vector is
      D: poly_vector;
   begin
      for i in 0 .. m-1 loop D(i) := 0; end loop;

      for i in 0 .. m-1 loop
         for k in 0 .. i loop
            D(i) := m2xor(D(i),m2and(a(k),b(i-k)));
         end loop;
      end loop;
   
      return D;
      
   end vector_D;

   
--------------------------------------------------------------------------------- 

   function vector_E(a,b: poly_vector) return poly_vector2 is
      E: poly_vector2;
   begin
      for i in 0 .. m-2 loop E(i) := 0; end loop;

      for i in 0 .. m-2 loop
         for k in i .. m-2 loop
            E(i) := m2xor(E(i),m2and(a(m-1-(k-i)),b(k+1)));
         end loop;
      end loop;
   
      return E;
      
   end vector_E;

--------------------------------------------------------------------------------- 

   function LSBfirst_squarer(A,F: poly_vector) return poly_vector is
      W,B,Faux, Baux: poly_vector;

   begin
      for i in 0 .. m-1 loop W(i) := 0; Baux(i) := 0; end loop;
      for i in 0 .. (m-2)/2 loop W(2*i) := A(i); end loop; -- even m
   
      B := F; -- even m
      Faux := Product_alpha(F);
   
      for k in 1 .. (m/2)-1 loop
         for i in 2 .. m-1 loop
            Baux(i) := m2xor(B(i-2),m2xor(m2and(B(m-1),Faux(i)),m2and(B(m-2),F(i))));
         end loop;
         Baux(1) := m2xor(m2and(B(m-1),Faux(1)),m2and(B(m-2),F(1)));
         Baux(0) := m2xor(m2and(B(m-1),Faux(0)),m2and(B(m-2),F(0)));
    
         W := m2xvv(W,m2abv(A((m/2)+k-1),B));
         B := Baux;
      end loop;   

      W := m2xvv(W,m2abv(A(m-1),B));

      return W;      

   end LSBfirst_squarer;

--------------------------------------------------------------------------------- 

   function LSBfirst(A,B,F: poly_vector) return poly_vector is
      Aux, C: poly_vector;

   begin
      for i in 0 .. m-1 loop C(i) := 0; end loop;
      Aux := A;   
      for i in 0 .. m-1 loop
         C := m2xvv(m2abv(B(i),Aux),C);
         Aux := Product_alpha_A(Aux,F);   
      end loop;   

   return C;
   
   end LSBfirst;

--------------------------------------------------------------------------------- 

   function bsquarer_montgomery(A,F: poly_vector) return poly_vector is
      C: poly2_vector;
      D: poly_vector;

   begin
      for i in 0 .. 2*m-2 loop C(i) := 0; end loop;
      for i in 0 .. m-1 loop C(2*i) := A(i); end loop;

      for i in 0 .. m-1 loop
         if C(0) = 1 then 
            C := m2xvv2(C,F);
            C(m) := m2xor(C(m),1);
         end if;
         C := lshift2(C);   
      end loop;   

      for i in 0 .. m-1 loop
         D(i) := C(i);
      end loop;

      return D;      

   end bsquarer_montgomery;

--------------------------------------------------------------------------------- 

   function bmult_montgomery(A,B,F: poly_vector) return poly_vector is
      C: poly_vector;

   begin
      for i in 0 .. m-1 loop C(i) := 0; end loop;

      for i in 0 .. m-1 loop
         C := m2xvv(C,m2abv(A(i),B));
         if C(0) = 1 then 
            C := m2xvv(C,m2abv(C(0),F));
            C := lshift(C);
            C(m-1) := 1;
         else
            C := lshift(C);   
         end if;   
      end loop;   
      
      return C;
      
   end bmult_montgomery;

--------------------------------------------------------------------------------- 



end GF2m;

