----------------------------------------------------------------
-- GF 2**m package (GF2m.ads)
-- 
-- Defines several funtions for finite fields operations
-- of Chapter 7 examples
----------------------------------------------------------------
with Gnat.Io; use Gnat.Io;
with finite_fields_GF2m; use finite_fields_GF2m;

package GF2m is

   m: constant natural := 8;
   
   type poly_vector is new Bit_Vector(0 .. m-1);
   type poly_vector2 is new Bit_Vector(0 .. m-2);
   type poly_vectorm is new Bit_Vector(0 .. m);
   type poly2_vector is new Bit_Vector(0 .. 2*m-2);
   type poly_matrix_m1m2 is array (0 .. m-1, 0 .. m-2) of Bit;
   type poly_matrix_m2m1 is array (0 .. m-2, 0 .. m-1) of Bit;
   type poly_matrix is array (0 .. m-1, 0 .. m-1) of Bit;
   
--   P: constant poly_vector := (1,0,0,1,0,1,0,1);
  
   function Product_alpha_A(A,F: poly_vector) return poly_vector;
   function Product_alpha_Am(A,F: poly_vectorm) return poly_vectorm;
   function Product_alpha(F: poly_vector) return poly_vector;
   function m2abv(x: bit; y: poly_vector) return poly_vector;
   function m2xvv(x: poly_vector; y: poly_vector) return poly_vector;
   function m2xvvm(x: poly_vectorm; y: poly_vectorm) return poly_vectorm;
   function m2xvv2(x: poly2_vector; y: poly_vector) return poly2_vector;
   function lshift(x: poly_vector) return poly_vector;
   function rshift(x: poly_vector) return poly_vector;
   function lshiftm(x: poly_vectorm) return poly_vectorm;   
   function rshiftm(x: poly_vectorm) return poly_vectorm;
   function lshift2(x: poly2_vector) return poly2_vector;
   function poly_multiplication(A,B: poly_vector) return poly2_vector;
   function reduction_matrix_R (f: poly_vector) return poly_matrix_m1m2;
   function matrix_P (f: poly_vector) return poly_matrix_m2m1;
   function mastrovito_matrix (a: poly_vector; P: poly_matrix_m2m1) return poly_matrix;
   function vector_D (a,b: poly_vector) return poly_vector;
   function vector_E (a,b: poly_vector) return poly_vector2;
   function LSBfirst_squarer(A,F: poly_vector) return poly_vector;
   function LSBfirst(A,B,F: poly_vector) return poly_vector;
   function bsquarer_montgomery(A,F: poly_vector) return poly_vector;
   function bmult_montgomery(A,B,F: poly_vector) return poly_vector;


end GF2m;

