----------------------------------------------------------------------------
-- Mastrovito Multiplier, Pentanomial (mastrovito_pentanom_multiplier.vhd)
--
-- Computes the polynomial multiplication mod f in GF(2**m)
-- The hardware is genenerate for a specific class 1 pentanomial 
-- Where F = x^m + x^k3 + x^k2 + x^k1 + 1
-- 0 < k1 < k2 < k3 =< m/2; k1 = k3-k2
-- 
----------------------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity mastrovito_pentanom_multiplication is -- Z283 + z12 + z7 + z5 + 1
generic( M: natural := 8; k3: natural := 4; k2: natural := 3; k1: natural := 1);
--generic( M: natural := 283; k3: natural := 12; k2: natural := 7; k1: natural := 5);
port (
  a, b: in std_logic_vector(M-1 downto 0);
  c: out std_logic_vector(M-1 downto 0)
);
end mastrovito_pentanom_multiplication;

architecture simple of mastrovito_pentanom_multiplication is
  signal D: std_logic_vector (M-1 downto 0);
  signal E: std_logic_vector (M-2 downto 0);
begin

  genD: process(a,b)
    variable di: std_logic_vector(M-1 downto 0);
  begin
  for i in 0 to M-1 loop
    di(i) := '0';
    for k in 0 to i loop
      di(i) := (di(i) xor (a(k) and b(i-k)));
    end loop;
  end loop;
  D <= di;
  end process genD;

  genE: process(a,b)
    variable ei: std_logic_vector(M-2 downto 0);
  begin
  for i in 0 to M-2 loop
    ei(i) := '0';
    for k in i to M-2 loop
      ei(i) := (ei(i) xor (a(M-1-(k-i)) and b(k+1)));
    end loop;
  end loop;
  E <= ei;
  end process genE;


 --Mastrovito multiplication, optimized for class 1 Pentanomials
  mastrovitoPentan: process(e,d)
    variable h, e0, e1, e01: std_logic_vector(M-1 downto 0);
  begin
  
    for j in 0 to k2-2 loop
      h(j) := e(j+m-k3) xor e(j+m-k2);
    end loop;
    for j in 0 to k1-2 loop
      e0(j) := e(j) xor h(j) xor e(j+m-k1);
    end loop;
    for j in k1-1 to k2-2 loop
      e0(j) := e(j) xor h(j);
    end loop;
    for j in k2-1 to k3-2 loop
      e0(j) := e(j) xor e(j+m-k3);
    end loop;
    for j in k3-1 to m-2 loop
      e0(j) := e(j);
    end loop;
    e0(m-1) := '0';
    for j in 0 to k1-1 loop
      e1(j) := '0';
    end loop;
    for j in k1 to m-1 loop
      e1(j) := e0(j-k1);
    end loop;

    for j in 0 to k1-1 loop
      e01(j) := e0(j);
    end loop;
    for j in k1 to m-k2-1 loop
      e01(j) := e0(j) xor e1(j);
    end loop; 
    for j in m-k2 to m-2 loop
      e01(j) := h(j+k2-m);
    end loop;
    e01(m-1) := e1(m-1);

    for j in 0 to m-1 loop
      if j < k2 then
        c(j) <= d(j) xor e01(j);
      else
        c(j) <= d(j) xor e01(j) xor e01(j-k2);
      end if;
    end loop;
  end process;
end simple;
