----------------------------------------------------------------------------
-- Mastrovito Multiplier, trinom (mastrovito_trinom_multiplier.vhd)
--
-- Computes the polynomial multiplication mod f in GF(2**m)
-- The hardware is genenerate for a specific f.
--
-- 
----------------------------------------------------------------------------
--library IEEE;
--use IEEE.std_logic_1164.all;
--use IEEE.std_logic_arith.all;
--use IEEE.std_logic_unsigned.all;
--package mastrovito_trinom_multiplier_parameters is
--  constant M: integer := 233;
--  constant K: integer := 74;
--    --constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
--end mastrovito_trinom_multiplier_parameters;
--
------------------------------------------------------------
-- trinom
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
--use work.mastrovito_trinom_multiplier_parameters.all;

entity mastrovito_trinom_multiplication is
generic( M: natural := 233; K: natural := 74);
port (
  a, b: in std_logic_vector(M-1 downto 0);
  c: out std_logic_vector(M-1 downto 0)
);
end mastrovito_trinom_multiplication;

architecture simple of mastrovito_trinom_multiplication is
  signal D: std_logic_vector (M-1 downto 0);
  signal E: std_logic_vector (M-2 downto 0);
begin

  genD: process(a,b)
    variable di: std_logic_vector(M-1 downto 0);
  begin
  for i in 0 to M-1 loop
    di(i) := '0';
    for k in 0 to i loop
      di(i) := (di(i) xor (a(k) and b(i-k)));
    end loop;
  end loop;
  D <= di;
  end process genD;

  genE: process(a,b)
    variable ei: std_logic_vector(M-2 downto 0);
  begin
  for i in 0 to M-2 loop
    ei(i) := '0';
    for k in i to M-2 loop
      ei(i) := (ei(i) xor (a(M-1-(k-i)) and b(k+1)));
    end loop;
  end loop;
  E <= ei;
  end process genE;

 --Mastrovito multiplication, optimized for trinomials
  mastrovitoTrin: process(e,d)
    variable h: std_logic_vector(M-1 downto 0) := (others => '0');
  begin
    for j in k-2 downto (2*k-m-1) loop
      if j >= 0 then
        h(j) := e(j) xor e(j+m-k);
      end if;
    end loop;
    for j in (2*k-m-2) downto 0 loop
      if j >= 0 then 
        h(j) := e(j) xor h(j+m-k);
      end if;
    end loop;
   
    for j in 0 to k-2 loop
      c(j) <= d(j) xor h(j);
    end loop;
    c(k-1) <= d(k-1) xor e(k-1);
    for j in k to (2*k-2) loop
      if j < m-1 then
      c(j) <= d(j) xor e(j) xor h(j-k);
      end if;
    end loop;
    for j in (2*k-1) to m-2 loop  
      c(j) <= d(j) xor e(j) xor e(j-k);
    end loop;
    c(m-1) <= d(m-1) xor h(m-k-1);
  end process;
end simple;
