----------------------------------------------------------------------------
-- Montgomery combinational Multiplier (montg_comb_mult.vhd)
--
-- 
-- 
-- Computes the polynomial multiplication x.y.r^-1 mod f in GF(2**m)
-- Implements a sequential cincuit

-- Defines 2 entities (montg_cell and montgomery_mult)
-- 
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package montgomery_mult_comb_package is
  constant M: integer := 32;
  --constant F: std_logic_vector(M-1 downto 0):= "00011011"; --for M=8 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"001B"; --for M=16 bits
  constant F: std_logic_vector(M-1 downto 0):= x"0101001B"; --for M=32 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"010100000101001B"; --for M=64 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0000000000000000010100000101001B"; --for M=128 bits
  --constant F: std_logic_vector(M-1 downto 0):= "000"&x"00000000000000000000000000000000000000C9"; --for M=163
  --constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
end montgomery_mult_comb_package;

-----------------------------------
--  Montgomery multiplier basic data_path
-----------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.montgomery_mult_comb_package.all;

entity montg_cell is
port (
  c: in std_logic_vector(M-1 downto 0);
  b: in std_logic_vector(M-1 downto 0);
  a_i: in std_logic;
  new_c: out std_logic_vector(M-1 downto 0)
);
end montg_cell;

architecture rtl of montg_cell is
signal prev_c0: std_logic;
begin

  prev_c0 <= c(0) xor (a_i and b(0));

  datapath: for i in 1 to M-1 generate
    new_c(i-1) <= c(i) xor (a_i and b(i)) xor (F(i) and prev_c0);
  end generate;
  new_c(M-1) <= prev_c0;

end rtl;

-----------------------------------
-- Montgomery multiplier
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.montgomery_mult_comb_package.all;

entity montgomery_comb_mult is
port (
  a, b: in std_logic_vector (M-1 downto 0);
  c: out std_logic_vector (M-1 downto 0)
);
end montgomery_comb_mult;

architecture rtl of montgomery_comb_mult is
  component montg_cell is
  port (
    C: in std_logic_vector(M-1 downto 0);
    B: in std_logic_vector(M-1 downto 0);
    a_i: in std_logic;
    new_c: out std_logic_vector(M-1 downto 0)
  );
  end component montg_cell;

  signal inic, shift_r, ce_c: std_logic;
  signal count: natural range 0 to M;
  type states is range 0 to 3;
  signal current_state: states;
  type conections is array (0 to M) of std_logic_vector (M-1 downto 0);
  signal cc: conections;

begin

  cc(0) <= (others => '0');
  forg: for i in 0 to M-1 generate
    cell: montg_cell port map (C => cc(i), B => B, a_i => a(i), new_c => cc(i+1));
  end generate;
  c <= cc(M);

end rtl;
