----------------------------------------------------------------------------
-- Montgomery combinational Multiplier (montg_comb_mult.vhd)
--
-- 
-- 
-- Computes the polynomial multiplication x.y.r^-1 mod f in GF(2**m)
-- Implements acombinational cincuit

-- Defines 2 entities (montg_cell and montgomery_mult)
-- 
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package montgomery_squarer_comb_package is
  constant M: integer := 8;
  constant F: std_logic_vector(M-1 downto 0):= "00011011"; --for M=8 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"001B"; --for M=16 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0101001B"; --for M=32 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"010100000101001B"; --for M=64 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0000000000000000010100000101001B"; --for M=128 bits
  --constant F: std_logic_vector(M-1 downto 0):= "000"&x"00000000000000000000000000000000000000C9"; --for M=163
  --constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
end montgomery_squarer_comb_package;

-----------------------------------
--  Montgomery squarer basic cell
-----------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.montgomery_squarer_comb_package.all;

entity montg_sq_c_cell is
port (
  c: in std_logic_vector(2*M-2 downto 0);
  new_c: out std_logic_vector(2*M-2 downto 0)
);
end montg_sq_c_cell;

architecture rtl of montg_sq_c_cell is
begin

  datapath: for i in 1 to M-1 generate
    new_c(i-1) <= c(i) xor (F(i) and c(0));
  end generate;
  new_c(M-1) <= c(0) xor c(M);
  new_c(2*M-2 downto M) <= '0' & c(2*M-2 downto M+1);

end rtl;

-----------------------------------
-- Montgomery squarer
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.montgomery_squarer_comb_package.all;

entity montgomery_comb_squarer is
port (
  a: in std_logic_vector (M-1 downto 0);
  c: out std_logic_vector (M-1 downto 0)
);
end montgomery_comb_squarer;

architecture rtl of montgomery_comb_squarer is
  component montg_sq_c_cell is
  port (
    c: in std_logic_vector(2*M-2 downto 0);
    new_c: out std_logic_vector(2*M-2 downto 0)
  );
  end component montg_sq_c_cell;

  type conections is array (0 to M) of std_logic_vector (2*M-2 downto 0);
  signal cc: conections;

begin

  cc(0)(0) <= a(0);
  gen_aa: for i in 1 to M-1 generate
    cc(0)(2*I-1)<= '0'; cc(0)(2*I) <= a(i); 
  end generate;
  
  forg: for i in 0 to M-1 generate
    cell: montg_sq_c_cell port map (C => cc(i), new_c => cc(i+1));
  end generate;
  c <= cc(M)(M-1 downto 0);

end rtl;
