--------------------------------------------------------------------------------
--  Instantiate Binary and Montgomey exponentiation (ch7)
--  Generates random vectors and comparte resutls 
--
--
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.ALL;
use ieee.math_real.all; -- for UNIFORM, TRUNC

USE std.textio.ALL;
use work.exponentiation_sq_mult_package.all;

ENTITY test_exponentiation IS
END test_exponentiation;

ARCHITECTURE behavior OF test_exponentiation IS 

  
  -- Component Declaration for the Unit Under Test (UUT2)
  COMPONENT exponentiation_sq_mult is
    port (
      A: in std_logic_vector (M-1 downto 0);
      E: in std_logic_vector (N-1 downto 0);
      clk, reset, start: in std_logic; 
      B: out std_logic_vector (M-1 downto 0);
      done: out std_logic
    );
  END COMPONENT;
  
  COMPONENT exponentiation_sq_mult_adv is
    port (
      A: in std_logic_vector (M-1 downto 0);
      E: in std_logic_vector (N-1 downto 0);
      clk, reset, start: in std_logic; 
      B: out std_logic_vector (M-1 downto 0);
      done: out std_logic
    );
  END COMPONENT;

  COMPONENT exponentiation_montgomery is
    port (
      A: in std_logic_vector (M-1 downto 0);
      E: in std_logic_vector (N-1 downto 0);
      clk, reset, start: in std_logic; 
      B: out std_logic_vector (M-1 downto 0);
      done: out std_logic
    );
  END COMPONENT;
  
  COMPONENT exponentiation_montgomery_adv is
    port (
      A: in std_logic_vector (M-1 downto 0);
      E: in std_logic_vector (N-1 downto 0);
      clk, reset, start: in std_logic; 
      B: out std_logic_vector (M-1 downto 0);
      done: out std_logic
    );
  END COMPONENT;

  -- Internal signals
  SIGNAL x, b, b_adv, b_montg, b_montg_adv:  std_logic_vector(M-1 downto 0) := (others=>'0');
  SIGNAL y:  std_logic_vector(N-1 downto 0) := (others=>'0');
  SIGNAL clk, reset, start, done, done_adv, done_montg, done_montg_adv: std_logic;

  constant DELAY : time := 100 ns;
  constant PERIOD : time := 200 ns;
  constant DUTY_CYCLE : real := 0.5;
  constant OFFSET : time := 0 ns;
  constant NUMBER_TESTS: natural := 10;

BEGIN

  -- Instantiate the Unit Under Test (UUT)
  uut1:exponentiation_sq_mult PORT MAP( A => x, E => y, clk => clk, reset => reset, start => start, B => B, done => done);
--  uut2:exponentiation_sq_mult_adv PORT MAP( A => x, E => y, clk => clk, reset => reset, start => start, B => B_adv, done => done_adv);
  uut3:exponentiation_montgomery PORT MAP( A => x, E => y, clk => clk, reset => reset, start => start, B => b_montg, done => done_montg);
  uut4:exponentiation_montgomery_adv PORT MAP( A => x, E => y, clk => clk, reset => reset, start => start, B => b_montg_adv, done => done_montg_adv);
 
  
  PROCESS -- clock process for clk
  BEGIN
  WAIT for OFFSET;
  CLOCK_LOOP : LOOP
     clk <= '0';
     WAIT FOR (PERIOD *(1.0 - DUTY_CYCLE));
     clk <= '1';
     WAIT FOR (PERIOD * DUTY_CYCLE);
  END LOOP CLOCK_LOOP;
  END PROCESS;

  tb_proc : PROCESS --generate values
    PROCEDURE gen_random(X : out std_logic_vector; w: natural; s1, s2: inout Natural) IS
      VARIABLE i_x, aux: integer;
      VARIABLE rand: real;
    BEGIN
        aux := W/16;
        for i in 1 to aux loop
          UNIFORM(s1, s2, rand);
          i_x := INTEGER(TRUNC(rand * real(2**16)));
          x(i*16-1 downto (i-1)*16) := CONV_STD_LOGIC_VECTOR (i_x, 16);
        end loop;
        UNIFORM(s1, s2, rand);
        i_x := INTEGER(TRUNC(rand * real(2**(w-aux*16))));
        x(w-1 downto aux*16) := CONV_STD_LOGIC_VECTOR (i_x, (w-aux*16));
    END PROCEDURE;
    
    VARIABLE TX_LOC : LINE;
    VARIABLE TX_STR : String(1 to 4096);
    VARIABLE seed1, seed2: positive; 
    VARIABLE cycles, max_cycles, min_cycles, total_cycles: integer := 0;
    VARIABLE avg_cycles: real;
    VARIABLE initial_time, final_time: time;
    VARIABLE xx: std_logic_vector (M-1 DownTo 0) ;
    VARIABLE yy: std_logic_vector (N-1 DownTo 0) ;

  BEGIN
    min_cycles:= 2**20;
    start <= '0'; reset <= '1';
    WAIT FOR PERIOD;
    reset <= '0';
    WAIT FOR PERIOD;
    
    for I in 1 to NUMBER_TESTS loop
      gen_random(xx, M, seed1, seed2);
      x <= xx;
      gen_random(yy, N, seed1, seed2);
      y <= yy;
      
      start <= '1'; initial_time := now;
      WAIT FOR PERIOD;
      start <= '0';
      --wait until ((done = '1') and (done_montg = '1'));
      wait until (done = '1');
      final_time := now;
      cycles := (final_time - initial_time)/PERIOD;
      total_cycles := total_cycles+cycles;
      --ASSERT (FALSE) REPORT "Number of Cycles: " & integer'image(cycles) & "  TotalCycles: " & integer'image(total_cycles) SEVERITY WARNING;
      if cycles > max_cycles then  max_cycles:= cycles; end if;
      if cycles < min_cycles then  min_cycles:= cycles; end if;

      WAIT FOR 2*PERIOD;

      IF ( b /= b_montg or b /= b_montg_adv or b/= b_adv) THEN 
      --IF ( b/= b_adv) THEN 
        write(TX_LOC,string'("ERROR!!! b=")); write(TX_LOC, b);
        write(TX_LOC,string'("/= b_montg=")); write(TX_LOC, b_montg);
        write(TX_LOC,string'(") using: ( A =")); write(TX_LOC, x);
        write(TX_LOC,string'(", E =")); write(TX_LOC, y);
        write(TX_LOC, string'(", F = 1")); write(TX_LOC, F);
        write(TX_LOC, string'(" )"));
        TX_STR(TX_LOC.all'range) := TX_LOC.all;
        Deallocate(TX_LOC);
        ASSERT (FALSE) REPORT TX_STR SEVERITY ERROR;
      END IF;  
 
    end loop;
    WAIT FOR DELAY;
    avg_cycles := real(total_cycles)/real(NUMBER_TESTS);
    ASSERT (FALSE) REPORT
    "Simulation successful!.  MinCycles: " & integer'image(min_cycles) &
    "  MaxCycles: " & integer'image(max_cycles) & "  TotalCycles: " & integer'image(total_cycles) &
    "  AvgCycles: " & real'image(avg_cycles)
    SEVERITY FAILURE;
  END PROCESS;

END;
